#!/usr/bin/python

"""Usage: python alvert.py [options] infile outfile

Options:
  -h print this help screen
  -i [format] specifies the format of the input file [default = autodetect]
  -o [format] specifies the format of the output file [default = phylip]
      acceptable format options:
        phylip   PHYLIP format
        phylipn  PHYLIP interleaved (just use phylip for input)
        nexus    NEXUS format
        nexusn   NEXUS interleaved (just use nexus for input)
        msf      GCG multiple sequence format
        clustal  CLUSTAL format
        fasta    Standard FASTA format
        div      divtime format (output only!)
        pir      PIR/NBRF format
        mega     MEGA format
        must     MUST ali format
        gde      GDE format
        selex    SELEX format (input only!)
  -f fix sequence names by removing Genbank-type codes from in front of the name
  -b [nsites] bootstrap the alignment (100 replicates, if r option is not used).
      To bootstrap the number of sites in the original alignment, use -1 for the
      nsites argument.  Note: only one of b, j, k options is allowed
  -j [nsites] jackknife the alignment (100 replicates, if r option is not used).
      To jackknife 50% of the sites in the original alignment, use -1 for the
      nsites argument.  To jackknife a percentage of the sites, enter a
      percentage value (e.g., -j 75%).  Note: only one of b, j, k options is
      allowed.
  -k [nsites] jackstrap the alignment (100 replicates, if r option is not used).
      To jackstrap a percentage of the sites, enter a percentage value (e.g.,
      -k 75%).  Note: only one of b, j, k options is allowed.
  -r [replicates] resample the alignment the specified number of times (phylip 
      output formats only -- may only be used in conjunction with j, k, or b 
      option)
  -m multiple file mode: runs alvert on all files with the specified extension
      If this option is used, the infile argument should specify the extension
      of the input files, and outfile the extension of output files.  Base
      filenames will be kept (formerly batch mode!).
  -c concatenation mode: runs alvert on all files with the specified extension
      If this option is used, the infile argument should specify the extension
      of the input files (to be concatenated), and the outfile the file to which
      the concatenated data set will be written
  -v verbose output
        
"""

import getopt, glob, sys
from seqtools import *

version = '11 November 2008'

def usage():
  sys.stderr.write(__doc__)
  sys.exit(1)

if __name__ == '__main__':
  
  print 'alvert.py, current version as of %s.' % version
  
  opts, args = getopt.getopt(sys.argv[1:], 'hi:o:fb:j:k:r:mcv', ())
  
  if len(args) != 2:
    usage()
    
  goodformats = ('auto', 'phylip', 'phylipn', 'nexus', 'nexusn', 'mega', 'msf', 'clustal',
                  'fasta', 'div', 'pir', 'must', 'gde', 'selex')
    
  informat = 'auto'
  outformat = 'phylip'
  batch = False
  fix = False
  verbosearg = False
  bootarg = jackarg = jsarg = False
  concat = False
  replicates = 100
  
  for opt in opts:
    if opt[0] == '-h':
      usage()
    if opt[0] == '-i':
      informat = opt[1]
    elif opt[0] == '-o':
      outformat = opt[1]
    elif opt[0] == '-b':
      bootarg = True
      try:
        nsites = int(opt[1])
      except ValueError:
        sys.stderr.write('Number of replicates must be a positive integer!\n\n')
        usage()  
    elif opt[0] == '-j':
      jackarg = True
      nsites = opt[1]
    elif opt[0] == '-k':
      jsarg = True
      nsites = opt[1]
    elif opt[0] == '-r':
      #bootarg = True
      try:
        replicates = int(opt[1])
        if replicates < 1:
          raise ValueError
      except ValueError:
        sys.stderr.write('Number of replicates must be a positive integer!\n\n')
        usage()  
    elif opt[0] == '-m':
      batch = True
    elif opt[0] == '-c':
      concat = True
    elif opt[0] == '-f':
      fix = True
    elif opt[0] == '-v':
      verbosearg = True
      
  infiles = []
  outfiles = []
  
  if informat not in goodformats or outformat not in goodformats:
    usage()
    
  if informat == 'must' or informat == 'pir':
    informat = 'fasta'
  elif informat == 'nexusn':
    informat = 'nexus'
  elif informat == 'phylipn':
    informat = 'phylip'
  elif informat == 'div':
    sys.stderr.write('I don\'t know how to READ divtime format!\n\n')
    usage()
  elif outformat == 'selex':
    sys.stderr.write('I don\'t know how to WRITE SELEX format!\n\n')
  
  if bootarg or jackarg or jsarg:
    if outformat == 'phylip':
      interleaved = False
    elif outformat == 'phylipn':
      interleaved = True
    else:
      usage()
      
  if (bootarg and jackarg) or (bootarg and jsarg) or (jackarg and jsarg):
    usage()
  
  if batch:
    infiles = glob.glob('*.%s' % args[0])
    infiles.sort()
    extlen = len(args[0]) # num of chars in extension
    for file in infiles:
      outfiles.append(file[:-extlen] + args[1]) 
  
  elif concat:
    infiles = glob.glob('*.%s' % args[0])
    infiles.sort()
    outfiles.append(args[1])

  else:
    infiles.append(args[0])
    outfiles.append(args[1])
    
  for i in range(len(outfiles)): # changed from len(infiles)... double-check
    
    if concat:
      allaligns = getaligns(infiles)
      align = catseq(allaligns, logfilename='%s.log' % outfiles[0])
      
    else:
      infile = infiles[i]
      align = getseqs(infile, informat, verbosearg)

    outfile = outfiles[i]
    if outfile in os.listdir('.'):
      os.remove(outfile)

    if fix:
      fixnames(align)
    
    if bootarg:
      for i in range(replicates):
        try:
          bootalign = bootstrap(align, nsites)
        except SeqError, se:
          sys.stderr.write(se.value + '\n')
          usage()
        writephy(bootalign, outfile, interleaved = interleaved, 
                  verbose = verbosearg, append = True)
        
    elif jackarg:
      if nsites == '-1':
        thisnsites = -1
      elif nsites[-1] == '%':
        percentage = float(nsites[:-1])/100
        thisnsites = int(align.length * percentage)
      else:
        thisnsites = int(nsites)
      for i in range(replicates):
        try:
          jackalign = jackknife(align, thisnsites)
        except SeqError, se:
          sys.stderr.write(se.value + '\n')
          usage()
        writephy(jackalign, outfile, interleaved = interleaved, 
                  verbose = verbosearg, append = True)
        
    elif jsarg:
      if nsites[-1] == '%':
        percentage = float(nsites[:-1])/100
        thisnsites = int(align.length * percentage)
      else:
        thisnsites = int(nsites)
        for i in range(replicates):
          try:
            jsalign = jackstrap(align, thisnsites)
          except SeqError, se:
            sys.stderr.write(se.value + '\n')
            usage() 
          writephy(jsalign, outfile, interleaved = interleaved, 
                    verbose = verbosearg, append = True)
        
    elif outformat == 'phylip':
      writephy(align, outfile, verbose = verbosearg)
    elif outformat == 'phylipn':
      writephy(align, outfile, interleaved = True, verbose = verbosearg)
    elif outformat == 'nexus':
      writenex(align, outfile, verbose = verbosearg)
    elif outformat == 'nexusn':
      writenex(align, outfile, interleaved = True, verbose = verbosearg)
    elif outformat == 'msf':
      writemsf(align, outfile, verbose = verbosearg)
    elif outformat == 'mega':
      writemega(align, outfile, verbose = verbosearg)
    elif outformat == 'clustal':
      writeclustal(align, outfile, verbose = verbosearg)
    elif outformat == 'pir' or outformat == 'must':
      writefasta(align, outfile, variation = outformat, verbose = verbosearg)
    elif outformat == 'fasta':
      writefasta(align, outfile, verbose = verbosearg)
    elif outformat == 'div':
      writefasta(align, outfile, variation = outformat, verbose = verbosearg)
    elif outformat == 'gde':
      writegde(align, outfile, verbose = verbosearg)
      
    
