#!/usr/local/bin/python

"""
tresolve.py
Produces all possible trees with given constraints.

Usage: python tresolve.py constraintfile outfile

The constraint file should contain 1 tree per line, in Newick format.  Any 
polytomies present in the tree will be resolved in all possible ways.
Results from all trees in the constraint file will be written to the same 
output file.

"""

import glob, os, sys
from treetools2 import *
from types import *

# used to give a unique name to each temp file
filecount = 0

def tupaintupb(tupa, tupb):
  """Return True iff the values in tupa lie between those in tupb (2-tuples)."""
  
  return tupa[0] > tupb[0] and tupa[1] < tupb[1]

def findbraces(st):
  """Return a list containing indices of pairs of parentheses in st.

  The list of indices contains 2-tuples in which the first value is
  the index of the opening parenthesis and the index of the closing
  is the second value.  Both curly braces and parentheses are found.
  If a parenthesis isn't paired, StackError is raised.
  
  Arguments:
    st - the string in which {} and () are located
    
  """
  
  parenstack = Stack()
  curlystack = Stack()
  
  pairs = []
  
  for i in range(len(st)):
    try: 
      if st[i] == '(':
        parenstack.push(i)
      elif st[i] == ')':
        pairs.append((parenstack.pop(), i))
      elif st[i] == '{':
        curlystack.push(i)
      elif st[i] == '}':
        pairs.append((curlystack.pop(), i))
        
    except StackError:
      sys.stderr.write('Tree is invalid! Unbalanced %s at position %d.\n' \
                       % (st[i], i))
      raise
      
  if not parenstack.isempty():
    raise StackError('Unbalanced tree string! %d too many (.\n' \
                     % parenstack.size())
  if not curlystack.isempty():
    raise StackError('Unbalanced tree string! %d too many {.\n' \
                     % curlystack.size())
    
  return pairs

def findoutgroup(tree):
  """Return a tuple containing the tree and a single-taxon outgroup.
  
  In order to find a single-taxon outgroup, rerooting of the tree may be
  necessary.

  Arguments:
    tree - an UnresolvedTree object

  """
  
  node = tree
  
  # find an external node to use as the new outgroup
  while node.isinternal():
    node = node.lastchild()
  
  outgroup = node
  
  # the outgroup's parent will be the new root
  root = node.parent
  
  # reroot the tree with the new root, if necessary
  root.reroot()
  
  if root != tree:
    print '\nTree has been rerooted!'
    root.davinprint()
    print 'Rerooted tree in Newick format:'
    print str(root).replace('{','(').replace('}', ')')
    
  root.removechild(outgroup)
    
  return (root, outgroup)


def buildtreetemplate(templstr):
  """Create a TreeTemplate from a properly formatted Newick string.
  
  The templstr argument is a string representing a tree in Newick format,
  but with {} instead of () surrounding any unresolved nodes (polytomies).
  Such a string can be created from a normal Newick-format string by 
  building an UnresolvedTree from the tree string, and calling str() on
  the UnresolvedTree instance.
  
  Arguments:
    templstr - a Newick-format tree (as described above)
  
  """
  if '{' not in templstr:
    return templstr

  # if templstring is surrounded by {},
  #if templstr[0] == '{':
  else:
    braces = findbraces(templstr)[:-1] # ignore outermost brace
    groups = []
    it = len(braces) - 1
    try:
      outer = braces[-1]
    except IndexError:
      return TreeTemplate(templstr[1:-1].split(','))

    strlen = len(templstr)
    
    if outer[1] != strlen - 2: # if there are ungrouped taxa at end
      #print 'templstr:', templstr
      #print 'templstr[outer[1] + 2: strlen - 1]'
      #print 'templstr[-1]:', templstr[-1]
      #print 'quitting.'
      #sys.exit(1)
      groups.extend(templstr[outer[1] + 2: strlen - 1].split(','))

    while it >= 0:
      if tupaintupb(braces[it], outer):
        it -= 1
        continue
      
      # add any taxa that lie between groups
      between = templstr[braces[it][1] + 2:outer[0] - 1]
      if len(between) != 0:
        groups.extend(between.split(','))
      
      outer = braces[it]
      it -= 1
      #if templstr[outer[0]] == '(':
      #  groups.append(templstr[outer[0]:outer[1] + 1])
      #else:
      #  groups.append(buildtreetemplate(templstr[outer[0]:outer[1] + 1]))
      groups.append(buildtreetemplate(templstr[outer[0]:outer[1] + 1]))

    if outer[0] != 1: # if there are ungrouped taxa at the begninning
      groups.extend(templstr[1:outer[0] - 1].split(','))
    
    return TreeTemplate(groups)
    
  #else: # templstr surrounded by ()
  #  if '{' not in templstr:
  #    return templstr
  #  else:
  #    braces = findbraces(templstr)[:-1]
  #    prefix = suffix = ''
  #    it = len(braces) - 1
  #
  #    while it >= 0:
  #      prefix = templstr[:braces[it][0]]
  #      suffix = templstr[braces[it][1] + 1:]
  #      if templstr[braces[it][0]] == '{':
  #        break
  #      it -= 1

  #    variable = templstr[braces[it][0]:braces[it][1] + 1]
      
  #    return TreeTemplate(buildtreetemplate(variable), prefix = prefix, suffix = suffix)

def scramblebuild(template, treefile, outgroup = None):
  """Build all possible bifurcating trees from a TreeTemplate.

  Arguments:
    template - a TreeTemplate used to specify constraints on the trees
    treefile - file to which trees are written
    outgroup - used to unroot trees
  
  """
  
  taxa = []
  files = []
  for i in range(len(template)):      

    if isinstance(template[i], TreeTemplate):
      templatefile = open('.temp%d.tmp' % filecount, 'w+')
      globals()['filecount'] += 1
      scramblebuild(template[i], templatefile)
      # replace the template with the file pointer
      template[i] = templatefile
      
    if isinstance(template[i], FileType):
      files.append(template[i])
    elif isinstance(template[i], StringType):
      taxa.append(template[i])
      
  scrambleassist(files, 0, taxa, template.prefix, template.suffix, treefile, outgroup)
  treefile.seek(0)

def scrambleassist(tempfiles, index, groups, prefix, suffix, outfile, outgroup):
  """Helper function for scramblebuild.

  Arguments:
    tempfiles - a list of temporary files containing all resolutions of a 
                particular group of taxa
    index - current index of tempfile of interest
    groups - a list of strings, representing a particular resolution of a group
    prefix - constant prefix for all trees
    suffix - constant suffix for all trees
    outfile - file to which trees are written
    outgroup - used to unroot trees
  
  """
  
  
  if len(tempfiles) == 0:
    for tree in buildalltrees(groups):
      if outgroup:
        outfile.write('%s%s,%s)%s\n' % (prefix, tree[:-1], outgroup, suffix))
      else:
        outfile.write('%s%s%s\n' % (prefix, tree, suffix))
    return

  for line in tempfiles[index]:
    # temporarily append a line from the tempfile as a taxon
    groups.append(line.strip())
    if index == len(tempfiles) - 1:
      for tree in buildalltrees(groups):
        if outgroup:
          outfile.write('%s%s,%s)%s\n' % (prefix, tree[:-1], outgroup, suffix))
        else:
          outfile.write('%s%s%s\n' % (prefix, tree, suffix))
    else:
      scrambleassist(tempfiles, index + 1, groups, prefix, suffix, outfile, outgroup)
    
    groups[-1:] = []
    
  tempfiles[index].seek(0)

def terminate(exitcode):
  """Delete temp files and terminate the program with the given exit code."""
  
  for tf in glob.glob('.temp*.tmp'):
    os.remove(tf)
  
  sys.exit(exitcode)
  
  
class TreeTemplate:
  def __init__(self, variable, prefix = '', suffix = ''):
    self.prefix = prefix
    self.suffix = suffix
    self.variable = variable
    
  def __getitem__(self, key):
    return self.variable[key]
  
  def __setitem__(self, key, value):
    self.variable[key] = value
  
  def __iter__(self):
    for group in self.variable:
      yield group
     
  def __len__(self):
    return len(self.variable)

  def __str__(self):
    return repr(self)
  
  def __repr__(self):
    s = self.prefix + '{'
    for group in self.variable:
      if isinstance(group, StringType):
        s += group + ','
      else:
        s += repr(group) + ','
    s = s[:-1] + '}' + self.suffix
    
    return s
    
  def template2tuple(self):
    return (self.prefix, self.suffix, self.variable)
  
class TemplateError(Exception):
  def __init__(self, message):
    self.value = message
      
  def __repr__(self):
    return repr(message)
    
  def __str__(self):
    return repr(self)  

  
class Stack:
  def __init__(self):
    self.list = []
  
  def push(self, elem):
    self.list.append(elem)
    
  def pop(self):
    if len(self.list) == 0:
      raise StackError('Can\'t pop from an empty stack!')

    elem = self.list[-1]
    self.list[-1:] = []
    
    return elem
  
  def peek(self):
    if len(self.list) == 0:
      raise StackError('Can\'t peek at an empty stack!')
    
    return self.list[-1]
  
  def size(self):
    return len(self.list)
  
  def isempty(self):
    return self.size() == 0
    
class StackError(Exception):
  def __init__(self, message):
    self.value = message
      
  def __repr__(self):
    return repr(message)
    
  def __str__(self):
    return repr(self)

if __name__ == '__main__':

  if len(sys.argv) != 3:
    sys.stderr.write(__doc__)
    terminate(1)
    
  infile = open(sys.argv[1])
  
  dlist = os.listdir('.')
  while sys.argv[2] in dlist:
    while True:
      answer = raw_input('A file named %s already exists.\n' % sys.argv[2]+ \
                         '(A)ppend to this file, (D)elete this file, (R)ename output file, or (Q)uit? ')
      answer = answer.lower()
      if answer == 'a':
        dlist.remove(sys.argv[2])
        break
      if answer == 'd':
        os.remove(sys.argv[2])
        dlist.remove(sys.argv[2])
        break
      elif answer == 'r':
        sys.argv[2] = raw_input('What would you like to call your output file? ')
        break
      elif answer == 'q':
        terminate(0)
      else:
        sys.stderr.write('Please answer \'a\', \'d\', \'r\', or \'q\'.\n')

  for line in infile:
    treestr = line.rstrip()
    print 'Testing syntax...',
    try:
      findbraces(treestr)
    except StackError:
      sys.stderr.write('Unbalanced parenthesis or curly brace in tree string!\n')
      terminate(1)
    print 'check.'
    
  
    tree = string2unresolvedtree(treestr)
    
    tree.davinprint()

    if not tree.isresolved():
      tree, outgroup = findoutgroup(tree)
      
    else:
      outgroup = None
      
    print 'tree:', tree
    print 'outgroup:', outgroup
    ntrees = tree.resolutions()
    if ntrees > 1:
      print 'The above tree will be resolved in all possible ways (%d trees). ' % ntrees
    else:
      raw_input('The above tree is already resolved, skipping (press enter).')  
      continue

    while True:
      answer = raw_input('Continue? (y/n) ').lower()
      if answer == 'y':
        break
      elif answer == 'n':
        terminate(0)
      else:
        print 'Please answer y or n.'
      
    treestr = str(tree)[:-1] # get rid of semi-colon
      
    template = buildtreetemplate(treestr)
    print 'template:'
    print template

    #if template.suffix == '':
    #  template.suffix = ';'
    template.suffix += ';'
    
    print 'Tresolving...'
    outfile = open(sys.argv[2], 'a')
    scramblebuild(template, outfile, outgroup)
    outfile.close()
    
  terminate(0)
